//
// Copyright (c) 2012, MindFusion LLC - Bulgaria.
//

using System;
using System.Drawing;
using System.Windows.Forms;

using Orientation = MindFusion.Diagramming.Orientation;
using SolidBrush = MindFusion.Drawing.SolidBrush;


namespace MindFusion.Diagramming.WinForms.Samples.CS.WorkflowDesigner
{
	public partial class MainForm : Form
	{
		public MainForm()
		{
			InitializeComponent();

			GlassEffect effect = new GlassEffect();
			effect.Type = GlassEffectType.Type4;
			effect.GlowColor = Color.Black;
			diagram.DiagramStyle.NodeEffects.Add(effect);
		}

		private void MainForm_Load(object sender, EventArgs e)
		{
			view.InplaceEditAcceptOnEnter = true;
			view.DelKeyAction = DelKeyAction.None;

			diagram.LinkHeadShape = ArrowHeads.Triangle;
			diagram.LinkHeadShapeSize = 3;
			diagram.DiagramLinkStyle.Brush = new SolidBrush(Color.LightGray);

			NewWorkflow();
		}

		private void btnNew_Click(object sender, EventArgs e)
		{
			NewWorkflow();
		}

		private void btnLayout_Click(object sender, EventArgs e)
		{
			LayoutDiagram();
		}

		private void btnSave_Click(object sender, EventArgs e)
		{
			SaveFileDialog sfd = new SaveFileDialog();
			if (sfd.ShowDialog() == DialogResult.OK)
				diagram.SaveToFile(sfd.FileName);
		}

		private void btnLoad_Click(object sender, EventArgs e)
		{
			OpenFileDialog ofd = new OpenFileDialog();
			if (ofd.ShowDialog() == DialogResult.OK)
				diagram.LoadFromFile(ofd.FileName);
		}

		#region Workflow creation

		private void NewWorkflow()
		{
			diagram.Tag = 0;
			diagram.ClearAll();

			ShapeNode start = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Start);
			start.Text = "Start";
			start.Brush = new SolidBrush(Color.LightGreen);
			start.LayoutTraits[WorkflowItemId] = "Start";
			start.LayoutTraits[PairStart] = diagram.Tag;

			ShapeNode stop = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Terminator);
			stop.Text = "Stop";
			stop.Brush = new SolidBrush(Color.Red);
			stop.LayoutTraits[PairEnd] = diagram.Tag;

			CreateLink(start, stop, true, "ControlFlow");

			view.Behavior = Behavior.Modify;

			LayoutDiagram();
			diagram.Tag = (int)diagram.Tag + 1;
		}

		private void CreateActivity(DiagramLink link)
		{
			ShapeNode origin = link.Origin as ShapeNode;
			ShapeNode dest = link.Destination as ShapeNode;

			string oldLogicID = link.LayoutTraits[Connector].ToString();
			diagram.Links.Remove(link);

			ShapeNode newActivity = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Process);
			newActivity.LayoutTraits[WorkflowItemId] = "Activity";
            newActivity.Brush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(119, 179, 238));

			CreateLink(origin, newActivity, true, oldLogicID);
			CreateLink(newActivity, dest, true, "ControlFlow");

			LayoutDiagram();

			view.BeginEdit(newActivity);
		}

		private void CreateIfElse(DiagramLink link)
		{
			ShapeNode origin = link.Origin as ShapeNode;
			ShapeNode dest = link.Destination as ShapeNode;

			string oldLogicID = link.LayoutTraits[Connector].ToString();
			diagram.Links.Remove(link);

			ShapeNode decision = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Decision);
			decision.Text = "Condition";
            decision.Brush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(156, 170, 198));
			decision.LayoutTraits[WorkflowItemId] = "IfElse";
			decision.LayoutTraits[PairStart] = diagram.Tag;

			ShapeNode endIf = diagram.Factory.CreateShapeNode(smallSize, Shapes.Ellipse);
			endIf.Text = "";
            endIf.Brush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(156, 170, 198));
			endIf.LayoutTraits[PairEnd] = diagram.Tag;

			CreateLink(origin, decision, true, oldLogicID);
			CreateLink(decision, endIf, true, "Yes");
			CreateLink(decision, endIf, true, "No");
			CreateLink(endIf, dest, true, "ControlFlow");

			LayoutDiagram();
			diagram.Tag = (int)diagram.Tag + 1;
		}

		private void CreateWhile(DiagramLink link)
		{
			ShapeNode origin = link.Origin as ShapeNode;
			ShapeNode dest = link.Destination as ShapeNode;

			string oldLogicID = link.LayoutTraits[Connector].ToString();
			diagram.Links.Remove(link);

			ShapeNode loopStart = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.BeginLoop);
            loopStart.Brush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(192, 192, 192));
			loopStart.LayoutTraits[WorkflowItemId] = "While";
			loopStart.Text = "while";
			loopStart.LayoutTraits[PairStart] = diagram.Tag;

			ShapeNode loopEnd = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.EndLoop);
            loopEnd.Brush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(192, 192, 192));
			loopEnd.Text = "end while";
			loopEnd.LayoutTraits[PairEnd] = diagram.Tag;

			CreateLink(origin, loopStart, true, oldLogicID);
			CreateLink(loopStart, loopEnd, true, "ControlFlow");
			CreateLink(loopEnd, dest, true, "ControlFlow");

			DiagramLink backLink = CreateLink(loopEnd, loopStart, false, "While");
			backLink.IgnoreLayout = true;

			LayoutDiagram();
			diagram.Tag = (int)diagram.Tag + 1;
		}

		private DiagramLink CreateLink(ShapeNode n1, ShapeNode n2,
			bool acceptActivities, string logicID)
		{
			DiagramLink link = diagram.Factory.CreateDiagramLink(n1, n2);
			link.Tag = acceptActivities;
			if (acceptActivities)
				link.Text = "drop here";

			link.LayoutTraits[Connector] = logicID;

			return link;
		}

		#endregion

		#region Workflow layout

		private void LayoutDiagram()
		{
			view.EndEdit(true);

			LayoutBranch(diagram.Nodes[0], 0);
			diagram.ResizeToFitItems(10);
		}

		private void LayoutBranch(DiagramNode node, int branchEnd)
		{
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return;

			DiagramNode continueFrom = null;

			int endMarker = node.LayoutTraits[PairStart] == null ?
				branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink trueLink = GetOutLink(node, "Yes");
				DiagramLink falseLink = GetOutLink(node, "No");
				continueFrom = GetPairEnd(endMarker);

				// layout the "true" branch at the left side under the decision node
				if (trueLink.Destination != continueFrom)
				{
					float rightWidth = GetRightWidth(trueLink.Destination, endMarker);
					DiagramNode branchStart = trueLink.Destination;
					branchStart.Bounds = new RectangleF(node.Bounds.Left +
						node.Bounds.Width / 2 - rightWidth - branchStart.Bounds.Width / 2 - 5,
						node.Bounds.Bottom + LevelDistance, branchStart.Bounds.Width, branchStart.Bounds.Height);
					LayoutBranch(branchStart, endMarker);
				}

				// layout the "false" branch at the right side under the decision node
				if (falseLink.Destination != continueFrom)
				{
					float leftWidth = GetLeftWidth(falseLink.Destination, endMarker);
					DiagramNode branchStart = falseLink.Destination;
					branchStart.Bounds = new RectangleF(node.Bounds.Right -
						node.Bounds.Width / 2 + leftWidth - branchStart.Bounds.Width / 2 + 5,
						node.Bounds.Bottom + LevelDistance, branchStart.Bounds.Width, branchStart.Bounds.Height);
					LayoutBranch(branchStart, endMarker);
				}

				// continue the layout below the longer branch
				RectangleF bb = continueFrom.Bounds;
				DiagramNode e1 = continueFrom.IncomingLinks[0].Origin;
				DiagramNode e2 = continueFrom.IncomingLinks[1].Origin;
				continueFrom.Bounds = new RectangleF(
					node.Bounds.Left + node.Bounds.Width / 2 - bb.Width / 2,
					Math.Max(e1.Bounds.Bottom, e2.Bounds.Bottom) + LevelDistance,
					bb.Width, bb.Height);

				// bend the "true" link
				if (trueLink.Destination != continueFrom)
					BendLinkLeftDown(trueLink);
				else
					BendLinkLeftDownRight(trueLink);

				// bend the "false" link
				if (falseLink.Destination != continueFrom)
					BendLinkRightDown(falseLink);
				else
					BendLinkRightDownLeft(falseLink);

				// bend the branch ending links
				for (int l = 0; l < 2; ++l)
				{
					if (continueFrom.IncomingLinks[l].Origin != node)
					{
						DiagramLink link = continueFrom.IncomingLinks[l];
						BendLinkDownLOR(link);
					}
				}
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				"Start".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				continueFrom = GetOutLink(node, "ControlFlow").Destination;
				RectangleF bb = continueFrom.Bounds;
				continueFrom.Bounds = new RectangleF(
					node.Bounds.Left + node.Bounds.Width / 2 - bb.Width / 2,
					node.Bounds.Bottom + LevelDistance,
					bb.Width, bb.Height);

				DiagramLink link = GetOutLink(node, "ControlFlow");
				SetLinkStraight(link);
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramNode branchStart = GetOutLink(node, "ControlFlow").Destination;
				RectangleF bb = branchStart.Bounds;
				branchStart.Bounds = new RectangleF(
					node.Bounds.Left + node.Bounds.Width / 2 - bb.Width / 2,
					node.Bounds.Bottom + LevelDistance,
					bb.Width, bb.Height);
				LayoutBranch(branchStart, endMarker);

				DiagramLink link = GetOutLink(node, "ControlFlow");
				SetLinkStraight(link);

				float rightWidth = GetRightWidth(branchStart, endMarker);
				continueFrom = GetPairEnd(endMarker);

				DiagramLink backLink = GetOutLink(continueFrom, "While");
				BendBackLink(backLink, rightWidth);
			}

			if (continueFrom != null)
				LayoutBranch(continueFrom, branchEnd);
		}

		private DiagramNode GetPairEnd(int pairId)
		{
			foreach (DiagramNode node in diagram.Nodes)
			{
				if (node.LayoutTraits[PairEnd] != null &&
					(int)node.LayoutTraits[PairEnd] == pairId)
					return node;
			}

			return null;
		}

		private DiagramLink GetOutLink(DiagramNode node, string id)
		{
			foreach (DiagramLink link in node.OutgoingLinks)
			{
				if (link.LayoutTraits[Connector] is string &&
					(string)link.LayoutTraits[Connector] == id)
					return link;
			}

			return null;
		}

		// should merge the next 3 methods into a single one,
		// and calculate the width according to a parameter

		private float GetBranchWidth(DiagramNode node, int branchEnd)
		{
			float branchWidth = node.Bounds.Width;
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return branchWidth;

			int endMarker = node.LayoutTraits[PairStart] == null ? branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink trueLink = GetOutLink(node, "Yes");
				DiagramLink falseLink = GetOutLink(node, "No");
				branchWidth = Math.Max(node.Bounds.Width, 10 +
					GetBranchWidth(trueLink.Destination, endMarker) +
						GetBranchWidth(falseLink.Destination, endMarker)) + 10;
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				branchWidth = Math.Max(node.Bounds.Width,
					GetBranchWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				branchWidth = 5 + Math.Max(node.Bounds.Width,
					GetBranchWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}

			if (node.LayoutTraits[PairStart] != null)
			{
				int branchEndIdx = (int)node.LayoutTraits[PairStart];
				if (branchEndIdx != branchEnd)
				{
					branchWidth = Math.Max(branchWidth,
						GetBranchWidth(GetPairEnd(branchEndIdx), branchEnd));
				}
			}

			return branchWidth;
		}

		private float GetLeftWidth(DiagramNode node, int branchEnd)
		{
			float branchWidth = node.Bounds.Width / 2;
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return branchWidth;

			int endMarker = node.LayoutTraits[PairStart] == null ? branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink trueLink = GetOutLink(node, "Yes");
				branchWidth = Math.Max(node.Bounds.Width / 2, 5 +
					GetBranchWidth(trueLink.Destination, endMarker)) + 5;
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				branchWidth = Math.Max(node.Bounds.Width / 2,
					GetLeftWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				branchWidth = Math.Max(node.Bounds.Width / 2,
					GetLeftWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}

			if (node.LayoutTraits[PairStart] != null)
			{
				int branchEndIdx = (int)node.LayoutTraits[PairStart];
				if (branchEndIdx != branchEnd)
				{
					branchWidth = Math.Max(branchWidth,
						GetLeftWidth(GetPairEnd(branchEndIdx), branchEnd));
				}
			}

			return branchWidth;
		}

		private float GetRightWidth(DiagramNode node, int branchEnd)
		{
			float branchWidth = node.Bounds.Width / 2;
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return branchWidth;

			int endMarker = node.LayoutTraits[PairStart] == null ? branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink falseLink = GetOutLink(node, "No");
				branchWidth = Math.Max(node.Bounds.Width / 2, 5 +
					GetBranchWidth(falseLink.Destination, endMarker)) + 5;
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				branchWidth = Math.Max(node.Bounds.Width / 2,
					GetRightWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				branchWidth = 5 + Math.Max(node.Bounds.Width / 2,
					GetRightWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}

			if (node.LayoutTraits[PairStart] != null)
			{
				int branchEndIdx = (int)node.LayoutTraits[PairStart];
				if (branchEndIdx != branchEnd)
				{
					branchWidth = Math.Max(branchWidth,
						GetRightWidth(GetPairEnd(branchEndIdx), branchEnd));
				}
			}

			return branchWidth;
		}

		private PointF GetLeftMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Left, node.Bounds.Top + node.Bounds.Height / 2);
		}

		private PointF GetRightMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Right, node.Bounds.Top + node.Bounds.Height / 2);
		}

		private PointF GetTopMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Left + node.Bounds.Width / 2, node.Bounds.Top);
		}

		private PointF GetBottomMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Left + node.Bounds.Width / 2, node.Bounds.Bottom);
		}

		private void BendLinkLeftDown(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Shape = LinkShape.Cascading;
			link.SegmentCount = 2;
			PointF pt1 = GetLeftMiddle(link.Origin);
			PointF pt2 = GetTopMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt2.X, pt1.Y);
			link.ControlPoints[2] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkLeftDownRight(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Shape = LinkShape.Cascading;
			link.SegmentCount = 3;
			PointF pt1 = GetLeftMiddle(link.Origin);
			PointF pt2 = GetLeftMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt1.X - 5, pt1.Y);
			link.ControlPoints[2] = new PointF(pt1.X - 5, pt2.Y);
			link.ControlPoints[3] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkRightDown(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Shape = LinkShape.Cascading;
			link.SegmentCount = 2;
			PointF pt1 = GetRightMiddle(link.Origin);
			PointF pt2 = GetTopMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt2.X, pt1.Y);
			link.ControlPoints[2] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkRightDownLeft(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Shape = LinkShape.Cascading;
			link.SegmentCount = 3;
			PointF pt1 = GetRightMiddle(link.Origin);
			PointF pt2 = GetRightMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt1.X + 5, pt1.Y);
			link.ControlPoints[2] = new PointF(pt1.X + 5, pt2.Y);
			link.ControlPoints[3] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkDownLOR(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Vertical;
			link.Shape = LinkShape.Cascading;
			link.SegmentCount = 2;
			PointF pt1 = GetBottomMiddle(link.Origin);
			PointF pt2 = pt1.X > GetBottomMiddle(link.Destination).X ?
				GetRightMiddle(link.Destination) : GetLeftMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt1.X, pt2.Y);
			link.ControlPoints[2] = pt2;
			link.UpdateFromPoints();
		}

		private void BendBackLink(DiagramLink link, float offset)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Shape = LinkShape.Cascading;
			link.SegmentCount = 3;
			PointF pt1 = GetRightMiddle(link.Origin);
			PointF pt2 = GetRightMiddle(link.Destination);
			float m = GetBottomMiddle(link.Destination).X;
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(m + offset + 5, pt1.Y);
			link.ControlPoints[2] = new PointF(m + offset + 5, pt2.Y);
			link.ControlPoints[3] = pt2;
			link.UpdateFromPoints();
		}

		private void SetLinkStraight(DiagramLink link)
		{
			link.Shape = LinkShape.Polyline;
			link.SegmentCount = 1;
			link.ControlPoints[0] = GetBottomMiddle(link.Origin);
			link.ControlPoints[1] = GetTopMiddle(link.Destination);
			link.UpdateFromPoints();
		}

		#endregion

		#region Drag and drop

		private void palette_MouseDown(object sender, MouseEventArgs e)
		{
			dragStartPoint = new Point(e.X, e.Y);
			dragItem = null;
		}

		private void palette_MouseMove(object sender, MouseEventArgs e)
		{
			if (palette.SelectedIndex == -1 || e.Button != MouseButtons.Left)
				return;

			if (dragItem == null)
				dragItem = palette.Items[palette.SelectedIndex];

			if (Math.Abs(dragStartPoint.X - e.X) > SystemInformation.DragSize.Width ||
				Math.Abs(dragStartPoint.Y - e.Y) > SystemInformation.DragSize.Height)
			{
				DoDragDrop(dragItem, DragDropEffects.Copy);
			}
		}

		private void view_DragOver(object sender, DragEventArgs e)
		{
			DiagramLink link = LinkFromDragEvent(e);
			if (link != null && (bool)link.Tag)
				e.Effect = DragDropEffects.Copy;
			else
				e.Effect = DragDropEffects.None;
		}

		private void view_DragDrop(object sender, DragEventArgs e)
		{
			e.Effect = DragDropEffects.None;

			DiagramLink link = LinkFromDragEvent(e);
			if (link != null && (bool)link.Tag)
			{
				e.Effect = DragDropEffects.Copy;

				string type = e.Data.GetData(typeof(string)).ToString();
				switch (type)
				{
				case "Activity":
					CreateActivity(link);
					break;
				case "IfElse":
					CreateIfElse(link);
					break;
				case "While":
					CreateWhile(link);
					break;
				}
			}
		}

		private DiagramLink LinkFromDragEvent(DragEventArgs e)
		{
			// DragEventArgs gives us the screen coordinates of a point
			// so convert them to the diagram logical coordinate system
			Point clientPoint = view.PointToClient(new Point(e.X, e.Y));
			PointF diagramPoint = view.ClientToDoc(clientPoint);

			return diagram.GetLinkAt(diagramPoint, 2);
		}

		#endregion


		#region Fields
		
		RectangleF defaultPosition = new RectangleF(0, 0, 35, 20);
		RectangleF smallSize = new RectangleF(0, 0, 5, 5);
		object dragItem = null;
		Point dragStartPoint = Point.Empty;

		#endregion

		#region Constants

		// layout traits keys
		private const string WorkflowItemId = "WF";
		private const string PairStart = "PS";
		private const string PairEnd = "PE";
		private const string Connector = "CN";

		// size constants
		private float LevelDistance = 26;

		#endregion
	}
}